@@ -37,11 +37,16 @@ typedef struct hostent HOSTENT;
 typedef HOSTENT *PHOSTENT;
 #endif
 
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 typedef struct sockaddr_in6 SOCKADDR_IN6;
-typedef SOCKADDR_IN6 PI_SA;
+typedef union un_pi_sa {
+	SOCKADDR_IN v4;
+	SOCKADDR_IN6 v6;
+} PI_SA;
 #else
-typedef SOCKADDR_IN PI_SA;
+typedef union un_pi_sa {
+	SOCKADDR_IN v4;
+} PI_SA;
 #endif
 
 #if !(defined(HAVE_GETHOSTBYNAME_R)||defined(HAVE_NSL_GETHOSTBYNAME_R)||defined(_WIN32_))
