@@ -67,6 +67,8 @@ config_init_config (config_t * conf)
 
 	conf->allowlocalusers = 0;
 	conf->users = NULL;
+	conf->interfaces = NULL;
+	conf->lastiface = NULL;
 	conf->chains = NULL;
 
 	conf->sumlog.useConnFile = NULL;
@@ -213,7 +215,7 @@ config_get_addrtype (const char *string)
 		return AF_INET;
 	if (strcmp (string, "ip") == 0)
 		return AF_INET;
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 	if (strcmp (string, "ipv6") == 0)
 		return AF_INET6;
 #endif
@@ -310,7 +312,9 @@ config_get_addr (int addrtype, const char *string)
 	int nmask;
 	switch (addrtype) {
 	case AF_INET:
-		ret = (unsigned char *) malloc (16);
+		ret = (unsigned char *) malloc (8); /* IP+mask */
+		if (ret == NULL)
+			config_parseerror ("Out of memory");
 
 		/* If the address has a /, it also has a bitmask */
 		bitmask = strchr (string, '/');
@@ -342,7 +346,7 @@ config_get_addr (int addrtype, const char *string)
 		memcpy (ret + 4, &mask, 4);
 		return ret;
 		break;
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 	case AF_INET6:
 /* FIXME: Add IPv6 filtration */
 		config_parseerror ("IPv6 addresses not currently supported");
@@ -366,7 +370,7 @@ config_ac_start (void *data, const char *name, const char **atts)
 	ac_nodes *node;
 	int i;
 	stack *stk = (stack *) data;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 	sprintf (szDebug, "Loading tag %s", name);
 	DEBUG_LOG (szDebug);
@@ -402,8 +406,8 @@ config_ac_start (void *data, const char *name, const char **atts)
 				}
 				handledatt = TRUE;
 				choice->source_addrtype = config_get_addrtype (atts[i + 1]);
-				choice->have_source_addrtype = 1;
-				choice->have_source_addrtype_inherited = 1;
+				choice->have_source_addrtype = HAVE_YES;
+				choice->have_source_addrtype_inherited = HAVE_YES;
 			}
 			if (strcmp (atts[i], "source_addr") == 0) {
 				handledatt = TRUE;
@@ -412,12 +416,12 @@ config_ac_start (void *data, const char *name, const char **atts)
 						("source_addrtype must preceed source_addr");
 				choice->source_addr =
 					config_get_addr (choice->source_addrtype, atts[i + 1]);
-				choice->have_source_addr = 1;
+				choice->have_source_addr = HAVE_YES;
 			}
 			if (strcmp (atts[i], "source_port") == 0) {
 				handledatt = TRUE;
 				choice->source_port = atoi (atts[i + 1]);
-				choice->have_source_port = 1;
+				choice->have_source_port = HAVE_YES;
 			}
 			if (!handledatt)
 				config_parseerror ("attribute not recognised");
@@ -482,7 +486,7 @@ config_ch_start (void *data, const char *name, const char **atts)
 	BOOL handledatt = FALSE;
 	int i;
 	config_t *conf = (config_t *) data;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 	sprintf (szDebug, "Loading tag %s", name);
 	DEBUG_LOG (szDebug);
@@ -554,7 +558,7 @@ config_fil_start (void *data, const char *name, const char **atts)
 	fil_nodes *node;
 	int i;
 	stack *stk = (stack *) data;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 	sprintf (szDebug, "Loading tag %s", name);
 	DEBUG_LOG (szDebug);
@@ -660,12 +664,12 @@ config_fil_start (void *data, const char *name, const char **atts)
 			if (strcmp (atts[i], "authscheme") == 0) {
 				handledatt = TRUE;
 				filt->authscheme = config_get_authmech (atts[i + 1]);
-				filt->have_authscheme = 1;
+				filt->have_authscheme = HAVE_YES;
 			}
 			if (strcmp (atts[i], "authsrc") == 0) {
 				handledatt = TRUE;
 				filt->authsrc = config_get_authsrc (atts[i + 1]);
-				filt->have_authsrc = 1;
+				filt->have_authsrc = HAVE_YES;
 			}
 			if (strcmp (atts[i], "dest_addrtype") == 0) {
 				if (filt->have_dest_addrtype_inherited) {
@@ -674,8 +678,8 @@ config_fil_start (void *data, const char *name, const char **atts)
 				}
 				handledatt = TRUE;
 				filt->dest_addrtype = config_get_addrtype (atts[i + 1]);
-				filt->have_dest_addrtype = 1;
-				filt->have_dest_addrtype_inherited = 1;
+				filt->have_dest_addrtype = HAVE_YES;
+				filt->have_dest_addrtype_inherited = HAVE_YES;
 			}
 			if (strcmp (atts[i], "dest_addr") == 0) {
 				handledatt = TRUE;
@@ -684,22 +688,22 @@ config_fil_start (void *data, const char *name, const char **atts)
 						("dest_addrtype must preceed dest_addr");
 				filt->dest_addr =
 					config_get_addr (filt->dest_addrtype, atts[i + 1]);
-				filt->have_dest_addr = 1;
+				filt->have_dest_addr = HAVE_YES;
 			}
 			if (strcmp (atts[i], "dest_port") == 0) {
 				handledatt = TRUE;
 				filt->dest_port = atoi (atts[i + 1]);
-				filt->have_dest_port = 1;
+				filt->have_dest_port = HAVE_YES;
 			}
 			if (strcmp (atts[i], "throttle") == 0) {
 				handledatt = TRUE;
 				filt->throttle = atoi (atts[i + 1]);
-				filt->have_throttle = 1;
+				filt->have_throttle = HAVE_YES;
 			}
 			if (strcmp (atts[i], "socksop") == 0) {
 				handledatt = TRUE;
 				filt->socksop = config_get_socksop (atts[i + 1]);
-				filt->have_socksop = 1;
+				filt->have_socksop = HAVE_YES;
 			}
 			if (strcmp (atts[i], "source_addrtype") == 0) {
 				if (filt->have_source_addrtype_inherited) {
@@ -708,8 +712,8 @@ config_fil_start (void *data, const char *name, const char **atts)
 				}
 				handledatt = TRUE;
 				filt->source_addrtype = config_get_addrtype (atts[i + 1]);
-				filt->have_source_addrtype = 1;
-				filt->have_source_addrtype_inherited = 1;
+				filt->have_source_addrtype = HAVE_YES;
+				filt->have_source_addrtype_inherited = HAVE_YES;
 			}
 			if (strcmp (atts[i], "source_addr") == 0) {
 				handledatt = TRUE;
@@ -718,23 +722,23 @@ config_fil_start (void *data, const char *name, const char **atts)
 						("source_addrtype must preceed source_addr");
 				filt->source_addr =
 					config_get_addr (filt->source_addrtype, atts[i + 1]);
-				filt->have_source_addr = 1;
+				filt->have_source_addr = HAVE_YES;
 			}
 			if (strcmp (atts[i], "source_port") == 0) {
 				handledatt = TRUE;
 				filt->source_port = atoi (atts[i + 1]);
-				filt->have_source_port = 1;
+				filt->have_source_port = HAVE_YES;
 			}
 			if (strcmp (atts[i], "user") == 0) {
 				handledatt = TRUE;
 				filt->user = (char *) malloc (strlen (atts[i + 1]) + 1);
 				strcpy (filt->user, atts[i + 1]);
-				filt->have_user = 1;
+				filt->have_user = HAVE_YES;
 			}
 			if (strcmp (atts[i], "version") == 0) {
 				handledatt = TRUE;
 				filt->version = atoi (atts[i + 1]);
-				filt->have_version = 1;
+				filt->have_version = HAVE_YES;
 			}
 			if (!handledatt)
 				config_parseerror ("attribute not recognised");
@@ -772,7 +776,7 @@ config_sl_start (void *data, const char *name, const char **atts)
 	config_t *conf = (config_t *) data;
 	BOOL handled = FALSE;
 	int i;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 	sprintf (szDebug, "Loading tag %s", name);
 	DEBUG_LOG (szDebug);
@@ -902,6 +906,23 @@ config_adduser (config_t * conf, const char *user, const char *pass)
 					   CLEANUP_KEY_FREE | CLEANUP_VALUE_FREE);
 }
 
+static void
+config_addinterface (config_t * conf, const char *addrtype,
+					 const char *interface)
+{
+	interface_t *thisinterface;
+	thisinterface = (interface_t *) malloc (sizeof (interface_t));
+	thisinterface->addrtype = config_get_addrtype (addrtype);
+	thisinterface->addr =
+		config_get_addr (thisinterface->addrtype, interface);
+	thisinterface->next = NULL;
+	if (conf->lastiface == NULL) {
+		conf->interfaces = thisinterface;
+	} else {
+		conf->lastiface->next = thisinterface;
+	}
+	conf->lastiface = thisinterface;
+}
 
 void
 config_xml_start (void *data, const char *name, const char **atts)
@@ -910,7 +931,7 @@ config_xml_start (void *data, const char *name, const char **atts)
 	int i;
 	BOOL handled = FALSE;
 	BOOL handledatt = FALSE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 	sprintf (szDebug, "Loading tag %s", name);
 	DEBUG_LOG (szDebug);
@@ -1004,6 +1025,23 @@ config_xml_start (void *data, const char *name, const char **atts)
 			}
 		}
 	}
+	if (strcmp (name, "route") == 0) {
+		const char *addrtype = NULL;
+		const char *addr = NULL;
+		handled = TRUE;
+		for (i = 0; atts[i] && atts[i + 1]; i += 2) {
+			if (strcmp (atts[i], "addrtype") == 0) {
+				addrtype = atts[i + 1];
+			}
+			if (strcmp (atts[i], "interface") == 0) {
+				addr = atts[i + 1];
+			}
+		}
+		if (addrtype && addr) {
+			handledatt = TRUE;
+			config_addinterface (conf, addrtype, addr);
+		}
+	}
 	if (strcmp (name, "throttle") == 0) {
 		handled = TRUE;
 		for (i = 0; atts[i] && atts[i + 1]; i += 2) {
@@ -1090,6 +1128,17 @@ loadconfig (config_t * currentconfig, const char *filename)
 	return TRUE;
 }
 
+static unsigned char *
+config_getaddr (int addrtype, unsigned char *addr)
+{
+	switch (addrtype) {
+	case AF_INET:
+		return addr + 4;
+		break;
+	}
+	return NULL;
+}
+
 /* loc: 0==source, 1==dest */
 static BOOL
 config_isinaddr (int addrtype, unsigned char *addr, conn_t * conn, int loc)
@@ -1114,7 +1163,7 @@ config_isinaddr (int addrtype, unsigned char *addr, conn_t * conn, int loc)
 		memcpy (&ip, addrtmp, 4);
 		ip = ntohl (ip);
 		nodemask = (unsigned long *) addr;
-		nodeip = (unsigned long *) (unsigned char *) (addr + 4);
+		nodeip = (unsigned long *) config_getaddr (addrtype, addr);
 		if ((ip & *nodemask) != (*nodeip & *nodemask)) {
 			return FALSE;
 		}
@@ -1250,7 +1299,7 @@ BOOL
 config_isallowed (config_t * conf, conn_t * conn, chain_t ** chain)
 {
 	BOOL ret;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 #endif
 	conn->throttle = conf->throttle;
@@ -1258,7 +1307,7 @@ config_isallowed (config_t * conf, conn_t * conn, chain_t ** chain)
 	if (ret != 3) {
 		*chain = NULL;
 	}
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "Filter reports: %i (chain %p)", ret, (void *) *chain);
 	DEBUG_LOG (szDebug);
 #endif
@@ -1268,6 +1317,26 @@ config_isallowed (config_t * conf, conn_t * conn, chain_t ** chain)
 	return FALSE;
 }
 
+BOOL
+config_getroute (config_t * conf, conn_t * conn, SOCKADDR * sa, sl_t len)
+{
+	interface_t *tmp;
+	for (tmp = conf->interfaces; tmp; tmp = tmp->next) {
+		if (config_isinaddr (tmp->addrtype, tmp->addr, conn, 1)) {
+			sa->sa_family = tmp->addrtype;
+			switch (tmp->addrtype) {
+			case AF_INET:
+				((SOCKADDR_IN *) sa)->sin_addr.s_addr =
+					htonl (*(unsigned long *)
+						   config_getaddr (tmp->addrtype, tmp->addr));
+				((SOCKADDR_IN *) sa)->sin_port = htons (0);
+				return TRUE;
+			}
+		}
+	}
+	return FALSE;
+}
+
 int
 config_getMaxbindwait (config_t * conf)
 {
@@ -1430,7 +1499,7 @@ config_reference (config_t * conf)
 void
 config_dereference (config_t * conf)
 {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 #endif
 	os_mutex_lock (&conf->lock);
@@ -1443,7 +1512,7 @@ config_dereference (config_t * conf)
 		 * nothing is using it, so it can't hurt. */
 		os_mutex_close (&conf->lock);
 		free (conf);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		sprintf (szDebug, "Last reference to config %p, now destroyed",
 				 (void *) conf);
 		DEBUG_LOG (szDebug);
