@@ -73,10 +73,16 @@ os_pipe (int *ends)
 	return pipe (ends);
 }
 
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 void
-os_debug_log (const char *filename, const char *msg)
+os_debug_log (const char *function, const char *filename, unsigned int line,
+			  const char *msg)
 {
-	fprintf (stderr, "%s: %s\n", filename, msg);
+	if (function) {
+		fprintf (stderr, "%12s:%4i (%24s): %s\n", filename, line, function,
+				 msg);
+	} else {
+		fprintf (stderr, "%12s:%4i: %s\n", filename, line, msg);
+	}
 }
 #endif
