@@ -48,7 +48,7 @@ socks5_resolve_ipv4 (conn_t * conn)
 	return TRUE;
 }
 
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 static BOOL
 socks5_resolve_ipv6 (conn_t * conn)
 {
@@ -72,7 +72,7 @@ socks5_resolve_name (conn_t * conn)
 	char temp[257];
 	char tmpPort[3];
 	unsigned short port;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 #endif
 
@@ -80,7 +80,7 @@ socks5_resolve_name (conn_t * conn)
 
 	/* getSlab null terminates */
 	conn_getSlab (conn, temp, len);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "name trying to connect to %s", temp);
 	DEBUG_LOG (szDebug);
 #endif
@@ -105,7 +105,7 @@ socks5_revres_ipv4 (conn_t * conn, SOCKADDR * sa)
 	return TRUE;
 }
 
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 static BOOL
 socks5_revres_ipv6 (conn_t * conn, SOCKADDR * sa)
 {
@@ -133,7 +133,7 @@ socks5_revresolve (conn_t * conn, SOCKADDR * sa, BOOL bReportFail)
 		case AF_INET:
 			return socks5_revres_ipv4 (conn, sa);
 			break;
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 		case AF_INET6:
 			return socks5_revres_ipv6 (conn, sa);
 			break;
@@ -195,7 +195,7 @@ socks5_sendResponse (conn_t * conn, char rep, SOCKADDR * sa)
 static BOOL
 socks5_resolve (conn_t * conn, unsigned char atyp, BOOL bReportFail)
 {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[100];
 #endif
 	switch (atyp) {
@@ -205,13 +205,13 @@ socks5_resolve (conn_t * conn, unsigned char atyp, BOOL bReportFail)
 	case 0x03:					/* Hostname */
 		return socks5_resolve_name (conn);
 		break;
-#ifdef WITH_IPV6
+#ifdef ENABLE_IPV6
 	case 0x04:					/* IPv6 */
 		return socks5_resolve_ipv6 (conn);
 		break;
 #endif
 	default:
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		sprintf (szDebug, "No resolver %x", atyp);
 		DEBUG_LOG (szDebug);
 #endif
@@ -246,13 +246,13 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 	char *addrstring1;
 	char *addrstring2;
 	chain_t *chain;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 #endif
 
 	tcpsock = tcpconn->s;
 
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	DEBUG_LOG ("Entering UDP forward");
 #endif
 
@@ -273,7 +273,7 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 			biggest = two;
 		biggest++;
 		/* Wait for message from either side */
-		select (biggest, &fds, NULL, NULL, NULL);
+		select ((int) biggest, &fds, NULL, NULL, NULL);
 		if (FD_ISSET (tcpsock, &fds)) {
 			/* We got something on TCP.  I read somewhere that
 			 * some vendors have been extending the spec here.
@@ -284,18 +284,18 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 			len = recv (tcpsock, szBuffer, sizeof (szBuffer) - 1, 0);
 			if (len <= 0) {
 				finished = TRUE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 				DEBUG_LOG ("TCP close");
 #endif
 			} else {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 				DEBUG_LOG ("TCP activity");
 #endif
 			}
 		}
 		if (FD_ISSET (one, &fds)) {
 			sl = sizeof (sa_local);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 			DEBUG_LOG ("Outbound UDP packet");
 #endif
 			/* FIXME: Overwrites local with each incoming packet */
@@ -318,18 +318,26 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 				 * TCP connection */
 				addrstring1 = ai_getAddressString (&tcpconn->source);
 				addrstring2 = ai_getAddressString (&udpconn.source);
-				diff = strcmp (addrstring1, addrstring2);
-#ifdef WITH_DEBUG
+				diff = 1;
+				if (addrstring1 && addrstring2) {
+					diff = strcmp (addrstring1, addrstring2);
+				}
+#ifdef ENABLE_DEBUG
 				if (diff) {
-					sprintf
-						(szDebug,
-						 "TCP source (%s) != UDP source (%s), dropping packet",
-						 addrstring1, addrstring2);
-					DEBUG_LOG (szDebug);
+					if (addrstring1 && addrstring2) {
+						sprintf
+							(szDebug,
+							 "TCP source (%s) != UDP source (%s), dropping packet",
+							 addrstring1, addrstring2);
+						DEBUG_LOG (szDebug);
+					} else {
+						DEBUG_LOG ("Dropping UDP packet due to low memory");
+					}
+
 				}
 #endif
-				free (addrstring1);
-				free (addrstring2);
+				if (addrstring1) free (addrstring1);
+				if (addrstring2) free (addrstring2);
 				if (diff)
 					goto barfedoutbound;
 
@@ -363,10 +371,13 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 				if (!tcpconn->dest.nulladdr) {
 					addrstring1 = ai_getAddressString (&tcpconn->dest);
 					addrstring2 = ai_getAddressString (&udpconn.dest);
-					diff = strcmp (addrstring1, addrstring2);
-					free (addrstring1);
-					free (addrstring2);
-#ifdef WITH_DEBUG
+					diff = 1;
+					if (addrstring1 && addrstring2) {
+						diff = strcmp (addrstring1, addrstring2);
+					}
+					if (addrstring1) free (addrstring1);
+					if (addrstring2) free (addrstring2);
+#ifdef ENABLE_DEBUG
 					if (diff)
 						DEBUG_LOG ("TCP dest != UDP dest");
 #endif
@@ -385,7 +396,7 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 				tmplen = udpconn.bufflen - udpconn.buff_upto;
 				/* FIXME: FILTERING!! */
 				sendto (two, tmpbuf, tmplen, 0, remoteaddr, addrlen);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 				sprintf (szDebug, "Send %i bytes local->remote", len);
 				DEBUG_LOG (szDebug);
 #endif
@@ -393,7 +404,7 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 		}
 	  barfedoutbound:
 		if (FD_ISSET (two, &fds)) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 			DEBUG_LOG ("inbound UDP packet");
 #endif
 			/* A message from the world, forward to the LAN */
@@ -417,9 +428,12 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 				if (!tcpconn->dest.nulladdr) {
 					addrstring1 = ai_getAddressString (&tcpconn->dest);
 					addrstring2 = ai_getAddressString (&udpconn.source);
-					diff = strcmp (addrstring1, addrstring2);
-					free (addrstring1);
-					free (addrstring2);
+					diff = 1;
+					if (addrstring1 && addrstring2) {
+						diff = strcmp (addrstring1, addrstring2);
+					} 
+					if (addrstring1) free (addrstring1);
+					if (addrstring2) free (addrstring2);
 					if (diff)
 						goto barfedinbound;
 				}
@@ -439,7 +453,7 @@ socks5_forwardUDPData (conn_t * tcpconn, SOCKET one, SOCKET two)
 				 * destination should be allowed to reach the
 				 * source.*/
 				sendto (one, tmpbuf, tmplen, 0, localaddr, localsize);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 				sprintf (szDebug, "Send %i bytes remote->local", len);
 				DEBUG_LOG (szDebug);
 #endif
@@ -459,24 +473,28 @@ socks5_cmd_udp (conn_t * conn)
 	sl_t soutlen;
 	SOCKADDR *addr;
 	sl_t addrlen;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[300];
 #endif
 
 	if (!ai_getSockaddr (&conn->dest, &addr, &addrlen))
 		return FALSE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "Socket family: %x", addr->sa_family);
 	DEBUG_LOG (szDebug);
 #endif
 	remotelisten = socket (addr->sa_family, SOCK_DGRAM, IPPROTO_UDP);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	{
 		char *dest;
 		dest = ai_getString (&conn->dest);
-		sprintf (szDebug, "Binding to %s", dest);
-		DEBUG_LOG (szDebug);
-		free (dest);
+		if (dest) {
+			sprintf (szDebug, "Binding to %s", dest);
+			DEBUG_LOG (szDebug);
+			free (dest);
+		} else {
+			DEBUG_LOG ("Insufficient memory to report bind address");
+		}
 	}
 #endif
 	/* FIXME: This hack listens on all interfaces on any port.
@@ -485,7 +503,7 @@ socks5_cmd_udp (conn_t * conn)
 	((SOCKADDR *) & sout)->sa_family = addr->sa_family;
 	soutlen = addrlen;
 	if (bind (remotelisten, (SOCKADDR *) & sout, soutlen) != 0) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Bind failed");
 #endif
 		closesocket (remotelisten);
@@ -497,7 +515,7 @@ socks5_cmd_udp (conn_t * conn)
 	free (addr);
 	soutlen = sizeof (sout);
 	if (getsockname (remotelisten, (SOCKADDR *) & sout, &soutlen) != 0) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("getsockname failed");
 #endif
 	}
@@ -529,12 +547,17 @@ static void
 socks5_notifyclient (conn_t * conn, ANCONN s, SOCKADDR * sout, sl_t soutlen)
 {
 	if (an_getsockname (s, sout, soutlen) == AN_ERROR_SUCCESS) {
-#ifdef WITH_DEBUG
+		/* Only IPv4 atm */
+		if (sout->sa_family == AF_INET
+			&& ((SOCKADDR_IN *) sout)->sin_addr.s_addr == 0) {
+			config_getroute (conn->conf, conn, sout, soutlen);
+		}
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Bind succeeded, interface known");
 #endif
 		socks5_sendResponse (conn, 0x00, sout);
 	} else {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Bind succeeded, interface unknown");
 #endif
 		socks5_sendResponse (conn, 0x00, NULL);
@@ -553,25 +576,41 @@ socks5_cmd_bind (conn_t * conn, chain_t * chain)
 	sl_t addrlen;
 	fd_set fds;
 	struct timeval tv;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[100];
 #endif
 
 	if (!ai_getSockaddr (&conn->dest, &addr, &addrlen))
 		return FALSE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "Socket family: %x", addr->sa_family);
 	DEBUG_LOG (szDebug);
 #endif
 	remote = an_new_connection ();
-	conn_setupchain (conn, remote, chain);
-#ifdef WITH_DEBUG
+	if (remote != NULL) {
+		conn_setupchain (conn, remote, chain);
+		soutlen = sizeof (sout);
+		if (config_getroute (conn->conf, conn, (SOCKADDR *) & sout, soutlen)) {
+			unsigned int type;
+			if (an_getproxytype (remote, &type) == AN_ERROR_SUCCESS) {
+				if (type == AN_SERV_DIRECT) {
+					an_direct_setorigin (remote, (SOCKADDR *) & sout,
+										 soutlen);
+				}
+			}
+		}
+	}
+#ifdef ENABLE_DEBUG
 	{
 		char *dest;
 		dest = ai_getString (&conn->dest);
-		sprintf (szDebug, "Binding to %s", dest);
-		DEBUG_LOG (szDebug);
-		free (dest);
+		if (dest) {
+			sprintf (szDebug, "Binding to %s", dest);
+			DEBUG_LOG (szDebug);
+			free (dest);
+		} else {
+			DEBUG_LOG ("Insufficient memory to report bind address");
+		}
 	}
 #endif
 	/* FIXME: This hack listens on all interfaces on any port.
@@ -579,7 +618,7 @@ socks5_cmd_bind (conn_t * conn, chain_t * chain)
 	if ((remote == NULL) ||
 		(an_bind_tosockaddr (remote, (SOCKADDR *) addr, addrlen) !=
 		 AN_ERROR_SUCCESS) || (an_listen (remote) != AN_ERROR_SUCCESS)) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Bind failed");
 #endif
 		an_close (remote);
@@ -598,11 +637,11 @@ socks5_cmd_bind (conn_t * conn, chain_t * chain)
 	top = AN_FD_SET (remote, &fds, top);
 	tv.tv_sec = config_getMaxbindwait (conn->conf);
 	tv.tv_usec = 0;
-	select (top + 1, &fds, NULL, NULL, &tv);
+	select ((int) top + 1, &fds, NULL, NULL, &tv);
 
 	if (AN_FD_ISSET (remote, &fds)) {
 		addrinfo_t ai;
-		int correct;
+		int incorrect = 1;
 		char *whoArrived;
 		char *whoShould;
 		/* Now to wait for incoming connection */
@@ -615,7 +654,7 @@ socks5_cmd_bind (conn_t * conn, chain_t * chain)
 			/* Whoa, accept() failed.  Better tell client */
 			an_close (remote);
 			an_destroy (remote);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 			DEBUG_LOG ("an_accept failed");
 #endif
 			socks5_sendResponse (conn, 0x01, NULL);
@@ -625,12 +664,16 @@ socks5_cmd_bind (conn_t * conn, chain_t * chain)
 		whoArrived = ai_getAddressString (&ai);
 		whoShould = ai_getAddressString (&conn->dest);
 		ai_close (&ai);
-		correct = strcmp (whoArrived, whoShould);
-		free (whoArrived);
-		free (whoShould);
-		if (correct != 0) {
+		if (whoArrived && whoShould) {
+			/* If there's not enough memory to allocate these,
+			   then the request is denied */
+			incorrect = strcmp (whoArrived, whoShould);
+		}
+		if (whoArrived) free (whoArrived);
+		if (whoShould) free (whoShould);
+		if (incorrect != 0) {
 			/* Somebody connected who shouldn't have. */
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 			DEBUG_LOG ("Connection not from the right host");
 #endif
 			socks5_sendResponse (conn, 0x01, NULL);
@@ -647,7 +690,7 @@ socks5_cmd_bind (conn_t * conn, chain_t * chain)
 		/* As with connect, start forwarding data. */
 		conn_forwardData (conn, remote);
 	} else {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("select returned but without a connection");
 #endif
 		free (addr);
@@ -670,30 +713,45 @@ socks5_cmd_conn (conn_t * conn, chain_t * chain)
 	SOCKADDR *addr;
 	sl_t addrlen;
 	sl_t soutlen;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[100];
 #endif
 
 	if (!ai_getSockaddr (&conn->dest, &addr, &addrlen))
 		return FALSE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "Socket family: %x", addr->sa_family);
 	DEBUG_LOG (szDebug);
 #endif
 	remote = an_new_connection ();
-	conn_setupchain (conn, remote, chain);
-#ifdef WITH_DEBUG
+	if (remote != NULL) {
+		conn_setupchain (conn, remote, chain);
+		soutlen = sizeof (sout);
+		if (config_getroute (conn->conf, conn, (SOCKADDR *) & sout, soutlen)) {
+			unsigned int type;
+			if (an_getproxytype (remote, &type) == AN_ERROR_SUCCESS) {
+				if (type == AN_SERV_DIRECT)
+					an_direct_setorigin (remote, (SOCKADDR *) & sout,
+										 soutlen);
+			}
+		}
+	}
+#ifdef ENABLE_DEBUG
 	{
 		char *dest;
 		dest = ai_getString (&conn->dest);
-		sprintf (szDebug, "Connecting to %s", dest);
-		DEBUG_LOG (szDebug);
-		free (dest);
+		if (dest) {
+			sprintf (szDebug, "Connecting to %s", dest);
+			DEBUG_LOG (szDebug);
+			free (dest);
+		} else {
+			DEBUG_LOG ("Insufficient memory to report connect address");
+		}
 	}
 #endif
 
 	if (an_connect_tosockaddr (remote, addr, addrlen) != AN_ERROR_SUCCESS) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Connection failed");
 #endif
 		an_close (remote);
@@ -706,12 +764,12 @@ socks5_cmd_conn (conn_t * conn, chain_t * chain)
 	soutlen = sizeof (sout);
 	if (an_getsockname (remote, (SOCKADDR *) & sout, soutlen) ==
 		AN_ERROR_SUCCESS) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Connection succeeded, interface known");
 #endif
 		socks5_sendResponse (conn, 0x00, (SOCKADDR *) & sout);
 	} else {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Connection succeeded, interface unknown");
 #endif
 		socks5_sendResponse (conn, 0x00, NULL);
@@ -729,20 +787,20 @@ socks5_initial (conn_t * conn)
 	unsigned char nMethods;
 	unsigned char chosenmethod;
 	unsigned char szData[260];
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[100];
 #endif
 
 	if (!conn_getChar (conn, &nMethods))
 		return FALSE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "Number of Methods: %i", nMethods);
 	DEBUG_LOG (szDebug);
 #endif
 	if (!conn_getSlab (conn, (char *) szData, nMethods))
 		return FALSE;
 	chosenmethod = config_choosemethod (conn->conf, conn, szData, nMethods);
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	sprintf (szDebug, "Chosen method %i", chosenmethod);
 	DEBUG_LOG (szDebug);
 #endif
@@ -778,7 +836,7 @@ socks5_request (conn_t * conn)
 	unsigned char atyp;
 	BOOL ret;
 	chain_t *chain;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	char szDebug[100];
 #endif
 
@@ -795,7 +853,7 @@ socks5_request (conn_t * conn)
 	if (!socks5_resolve (conn, atyp, TRUE))
 		return FALSE;
 	if (conn->dest.addrlen == 0) {
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		DEBUG_LOG ("Resolver failed");
 #endif
 		socks5_sendResponse (conn, 0x04, NULL);	/* Host unreachable */
@@ -819,7 +877,7 @@ socks5_request (conn_t * conn)
 		ret = socks5_cmd_ident (conn);
 		break;
 	default:
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 		sprintf (szDebug, "No command %x", cmd);
 		DEBUG_LOG (szDebug);
 #endif
@@ -835,7 +893,7 @@ socks5_handler (conn_t * conn)
 {
 	if (!socks5_initial ((conn_t *) conn))
 		return FALSE;
-#ifdef WITH_DEBUG
+#ifdef ENABLE_DEBUG
 	DEBUG_LOG ("passed initialization");
 #endif
 	return socks5_request ((conn_t *) conn);
